/*
 * Copyright 2010 Semafor Informatik & Energie AG, Basel, Switzerland
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.

 */
//
// http://javaboutique.internet.com/tutorials/ehcache/
//
package ch.semafor.gendas.interceptors;

import java.io.Serializable;
import net.sf.ehcache.Cache;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Element;
import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Add Object Caching Using Spring, AOP, and Ehcache
 * @author Jacob Briscoe
 *
 */
public class MethodCachingInterceptor implements MethodInterceptor {

		private CacheManager cacheManager;
		private final Logger log = LoggerFactory.getLogger(getClass());
		private final String CACHE_DATA_KEY = "BEAN_LOADER";

		public MethodCachingInterceptor(){
			cacheManager=CacheManager.create();
		}
		
		public Object invoke(final MethodInvocation methodInvocation)
				throws Throwable {
			final String targetMethodName = methodInvocation.getMethod().getName();

			if (log.isInfoEnabled()) {
				log.info("Attempting cacheManager.getCache first run (no args lookup)");
			}

			Cache cache = cacheManager.getCache(targetMethodName);

			if (cache == null) { // Lets try with args now.
				if (log.isInfoEnabled()) {
					log.info("No cache found in no args lookup.");
				}
				final Object[] methodArgs = methodInvocation.getArguments();
				if (methodArgs != null) {
					final StringBuffer cacheName = new StringBuffer(targetMethodName);
					cacheName.append("(");
					for (int i = 0; i < methodArgs.length; i++) {
						cacheName.append(methodArgs[i].toString());
						if (i + 1 != methodArgs.length) {
							cacheName.append(",");
						}
					}
					cacheName.append(")");
					if (log.isInfoEnabled()) {
						log.info("Attempting cacheManager.getCache with args lookup:"
								+ cacheName);
					}
					cache = cacheManager.getCache(cacheName.toString());
				}
			}

			boolean methodInvocationProceed = false;
			boolean methodInvocationCache = false;

			Object methodReturn = null;

			if (cache != null) {
				if (log.isInfoEnabled()) {
					log.info("Cache Config. Found: " + cache.getName());
				}
				final Element cacheElement = cache.get(CACHE_DATA_KEY);
				if (cacheElement == null) {
					methodInvocationProceed = true;
				} else {
					if (log.isInfoEnabled()) {
						log.info("Cache Element Found");
					}
					methodReturn = cacheElement.getValue();
					if (log.isInfoEnabled()) {
						log.info("Using Cached Element for methodReturn.");
					}
				}
				methodInvocationCache = true;
			}

			if (cache == null || methodInvocationProceed) {
				methodReturn = methodInvocation.proceed();
				if (methodInvocationCache) {
					final Element newCacheElement = new Element(CACHE_DATA_KEY,
							(Serializable) methodReturn);
					cache.put(newCacheElement);
					if (log.isInfoEnabled()) {
						log.info("Created new CacheElement entry and stored in cache.");
					}
				}
			}

			return methodReturn;
		}

		/**
		 * @return the cacheManager
		 */
		public CacheManager getCacheManager() {
			return cacheManager;
		}

		/**
		 * @param cacheManager
		 *          the cacheManager to set
		 */
		public void setCacheManager(CacheManager cacheManager) {
			this.cacheManager = cacheManager;
		}

	}

