/*
 * Copyright 2010 Semafor Informatik & Energie AG, Basel, Switzerland
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.

 */
package ch.semafor.gendas.model;

import java.io.Serializable;
import java.util.Date;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.persistence.CascadeType;
import javax.persistence.Entity;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.NamedQueries;
import javax.persistence.NamedQuery;
import javax.persistence.OneToMany;
import javax.persistence.Table;
import javax.persistence.Transient;
import javax.persistence.Version;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.lang.builder.ToStringStyle;
import org.hibernate.annotations.Cache;
import org.hibernate.annotations.CacheConcurrencyStrategy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.semafor.gendas.dao.ElementDao;

@NamedQueries( {
    @NamedQuery(name = "Element.findByType", 
        query = "select e from Element e where e.elementType.name=:type "
        	+ "and 9999999 in (select nextRevision from e.modifications)"),
    @NamedQuery(name = "Element.findByTypeAndReference", 
        query = "select r.parent.parent from ElementRefList r "
        + "where r.parent.parent.elementType.name=:type "
        + "and :ref in elements(r.elementlist) "
        + "and r.nextRevision=9999999 "
        + "and 9999999 = (select max(nextRevision) from r.parent.parent.modifications)"),
//        + "order by r.parent.id asc"),
        @NamedQuery(name = "Element.findByTypeAndPropertyTypeAndStringLike", 
            query = "select pv.valuelist.property.element from PropertyValue pv "
            + "where lower(pv.svalue) like :v and "
            + "pv.valuelist.nextRevision=9999999 and "
            + "pv.valuelist.property.type.name=:ptype and "
            + "pv.valuelist.property.element.elementType.name=:etype "
            + "and 9999999 = (select max(nextRevision) from pv.valuelist.property.element.modifications)"),
    @NamedQuery(name = "Element.getCountByTypeAndPropertyTypeAndStringLike", 
        query = "select count(*) from PropertyValue pv "
        + "where lower(pv.svalue) like :v and "
        + "pv.valuelist.nextRevision=9999999 and "
        + "pv.valuelist.property.type.name=:ptype and "
        + "pv.valuelist.property.element.elementType.name=:etype "
        + "and 9999999 = (select max(nextRevision) from pv.valuelist.property.element.modifications)"),
    @NamedQuery(name = "Element.getCountByTypeAndJoin2PropertyTypesAndStringsLike", 
        query = "select count(*) from PropertyValue pv1,PropertyValue pv2  "
              + "where " +
                  "lower(pv1.svalue) like :v1 and "
              + "pv1.valuelist.property.type.name=:ptype1 and "
              + "lower(pv2.svalue) like :v2 and "
              + "pv2.valuelist.property.type.name=:ptype2 and "
              + "pv1.valuelist.nextRevision=9999999 and "
              + "pv2.valuelist.nextRevision=9999999 and "
              + "pv1.valuelist.property.element.elementType.name=:etype and "
              + "pv1.valuelist.property.element=pv2.valuelist.property.element "
              + "and 9999999 = (select max(nextRevision) from pv1.valuelist.property.element.modifications) "
              + "and 9999999 = (select max(nextRevision) from pv2.valuelist.property.element.modifications)"),
    @NamedQuery(name = "Element.getCountByTypeAndUnion2PropertyTypesAndStringsLike", 
        query = "select count(distinct pv1.valuelist.property.element) from PropertyValue pv1 "
              + "where " +
                  "lower(pv1.svalue) like :v1 and "
              + "pv1.valuelist.property.type.name=:ptype1 or "
              + "lower(pv1.svalue) like :v2 and "
              + "pv1.valuelist.property.type.name=:ptype2 and "
              + "pv1.valuelist.nextRevision=9999999 and "
              + "pv1.valuelist.property.element.elementType.name=:etype "
              + "and 9999999 = (select max(nextRevision) from pv1.valuelist.property.element.modifications) ")})
@Entity
@Table(name = "elements")
@Cache(usage = CacheConcurrencyStrategy.READ_WRITE)
public class Element implements Serializable {	
  @Transient
  private final Logger logger = LoggerFactory.getLogger(Element.class);
  private static final long serialVersionUID = 1L;

  @Id
  @GeneratedValue(strategy = GenerationType.AUTO)
  private Long id; // NOPMD by wim on 9/20/10 1:40 PM

  @Version
  private Long version;

  @OneToMany(mappedBy = "element", cascade = CascadeType.ALL)
  private List<Modification> modifications = new ArrayList<Modification>(); // NOPMD by wim on 9/20/10 1:47 PM

  @ManyToOne
  @JoinColumn(name = "element_type_id", nullable = false)
  private ElementType elementType;

  @OneToMany(mappedBy = "element", cascade = CascadeType.ALL)
  // @org.hibernate.annotations.Cascade(value = {
  // org.hibernate.annotations.CascadeType.DELETE_ORPHAN })
  private List<Property> properties = new ArrayList<Property>();

  @OneToMany(mappedBy = "parent", cascade = CascadeType.ALL)
  // @org.hibernate.annotations.Cascade(value = {
  // org.hibernate.annotations.CascadeType.DELETE_ORPHAN })
  private List<ElementRefs> references = new ArrayList<ElementRefs>(); 

  
  //flag to prevent multiple loads
  @Transient
  // ToDo new name, PMD: AvoidFieldNameMatchingMethodName
  private boolean isLoaded = false; 
    
  // prevent stack overflow when printing
  @Transient
  private boolean printing = false; 
  
  //prevent stack overflow when crunching
  @Transient
  private boolean crunched = false; 

  // Default Constructor
  public Element() throws CoreException{
    addModification();
  }

  // Constructor
  public Element(final ElementType type) throws CoreException {
    super();
    this.elementType = type;
    addModification();
  }

  public void setProperties(final List<Property> properties) {
    this.properties = properties;
  }

  public List<Property> getProperties() {
    return this.properties;
  }
  
  public List<ElementRefs> getReferences() {
    return this.references;
  }

  // getter and setter of Id
  public Long getId() {
    return this.id;
  }

  public void setId(final Long id) { // NOPMD by wim on 9/20/10 1:40 PM
    this.id = id;
  }

  public boolean isTransient() {
    return this.id == null;
  }

  public boolean isPersistent() {
    return this.id != null;
  }

  private Modification addModification() throws CoreException {
	  final Modification mod = new Modification(this);
    this.modifications.add(mod);
    return mod;
  }

  public List<Modification> getModifications() {
    return modifications;
  }

  // Get the last modification
  public Modification getLastModification() throws CoreException {
    if (modifications.isEmpty()) {
      return null; // NOPMD by wim on 9/20/10 2:50 PM
    }
    final int inx = modifications.size() - 1;
    final Modification hist = modifications.get(inx);
    if (hist.getNextRevision() != Modification.MaxRevision) {
      throw new CoreException("Last modification has not last revision number");
    }
    return hist;
  }

  // Get the new modification or create one if it does not exist.
  private Modification getNewModification() throws CoreException {
	  final Modification last_mod = getLastModification();
    if (last_mod != null) {
      if (last_mod.isTransient()) { // NOPMD by wim on 9/20/10 1:47 PM
        return last_mod; // NOPMD by wim on 9/20/10 2:50 PM
      }
    }
    final Modification new_mod = addModification();
    if (last_mod != null) {
      last_mod.setNextRevision(new_mod.getRevision());
    }
    return new_mod;
  }

  public Modification getModification( final long rev ){
    int inx = modifications.size() - 1;
    Modification hist;
    while( inx >= 0 ){
      hist = modifications.get(inx);
      if( hist.getRevision() == rev ){
        return hist; // NOPMD by wim on 9/20/10 2:50 PM
      }
      inx--;
    }
    return null;
  }

  public Modification getModification( final Date dat ){
    int inx = modifications.size() - 1;
    Modification hist;
    while( inx >= 0 ){
      hist = modifications.get(inx);
      if( dat.compareTo( hist.getTimestamp() ) >= 0 ){
        return hist; // NOPMD by wim on 9/20/10 2:50 PM
      }
      inx--;
    }
    return null;
  }

  // Get the last Revisionnumber
  public long getLastRevision() throws CoreException {
	final Modification last_mod = getLastModification();
    if (last_mod == null) {
      throw new CoreException("An Element without a modification is not possible");
    }
    return last_mod.getRevision();
  }

  // Get the new Revisionnumber and create a transient new modification
  // if it does not exist.
  public long getNewRevision() throws CoreException {
	final Modification hist = getNewModification();
    if (hist == null) {
      throw new CoreException("New modification not created");
    }
    return hist.getRevision();
  }

  // getter and setter of ElementType
  public ElementType getElementType() {
    return this.elementType;
  }

  public void setElementType(final ElementType elementType) {
    this.elementType = elementType;
  }

  // get Property by type
  public Property getProperty(PropertyType t) { // NOPMD by wim on 9/20/10 1:40 PM
    for (Property p : this.properties) {
      if (t.equals(p.getType())) {
        return p; // NOPMD by wim on 9/20/10 2:50 PM
      }
    }
    return null;
  }

  // get Property by name
  public Property getProperty(final String name) {
    for (Property p : this.properties) {
      if (name.equals(p.getType().getName())) {
        return p;
      }
    }
    return null;
  }

  // add a new Property
  public void addProperty(final Property prop) throws CoreException {
    if (!this.equals(prop.getElement())) {
      throw new CoreException("Property not attached to this Element");
    }
    if (properties.contains(prop)) {
      throw new CoreException("Property already added to Element");
    }
    this.properties.add(prop);
  }

  // add a new ElementRefs
  public void addElementRefs(ElementRefs r) throws CoreException { // NOPMD by wim on 9/20/10 1:40 PM
    if (getElementRefs(r.getRefName()) != null) {
      throw new CoreException("ElementRefs '" + r.getRefName()
          + "' already exists");
    }
    r.setParent(this);
    r.setRevision(getNewRevision());
    r.setNextRevision(Modification.MaxRevision);
    r.crunch();
    this.references.add(r);
  }

  // get ElementRefs by name
  public ElementRefs getElementRefs(final String key) {
    for (ElementRefs r : this.references) {
      if (r.getRefName().equals(key)) {
        return r;
      }
    }
    return null;
  }

  // get a list of all Elementrefs
  public List<ElementRefs> getListOfElementRefs() {
    return this.references;
  }

  public ElementRefList getElementRefList(final String key, final long rev) {
	final  ElementRefs r = getElementRefs(key); // NOPMD by wim on 9/20/10 1:42 PM
    if (r != null) {
      if (r.isInRevision(rev)) { // NOPMD by wim on 9/20/10 1:47 PM
        return r.getElementRefList(rev);
      }
    }
    return null;
  }

  public ElementRefList getLastElementRefList(final String key) {
    return getElementRefList(key, Modification.MaxRevision);
  }

  public void setListOfElements(final String key, final List<Element> ellist)
      throws CoreException {
      logger.debug("Element::setListOfElements(): key={}", key);
    ElementRefs r = getElementRefs(key); // NOPMD by wim on 9/20/10 1:42 PM
    if (r != null) {
      r.setListOfElements(ellist);
      return;
    }
    // element ref does not yet exist
    if (ellist == null || ellist.size() == 0){ // NOPMD this JDK 1.5 by wim on 9/20/10 1:44 PM
      return;
    }
    r = new ElementRefs(this, key);
    r.setListOfElements(ellist);
  }

  public void addElement(String key, Element el) throws CoreException { // NOPMD by wim on 9/20/10 1:42 PM
      logger.debug("Element::addReference(): key= {}", key);
    ElementRefs r = getElementRefs(key); 
    if (r != null) {
      r.addElement(el);
      return;
    }
    // Not yet created
    r = new ElementRefs(this, key);
    r.addElement(el);
  }

  public List<Element> getListOfElements(final String key) throws CoreException {
	final ElementRefs r = getElementRefs(key); // NOPMD by wim on 9/20/10 1:42 PM
    if (r == null) {
      return null;
    }
    return r.getLastListOfElements();
  }

  // Comparison of two Elements
  public boolean equals(Object o) { // NOPMD by wim on 9/20/10 1:45 PM
    if (this == o) {
      return true;
    }
    if (!(o instanceof Element)) {
      return false;
    }
    final Element e = (Element) o; // NOPMD by wim on 9/20/10 1:45 PM
    if (this.elementType != e.elementType) {
      return false;
    }
    // TODO: what exactly is equality ? 
    if( this.id != null && e.id!=null ){
      if( this.id.equals(e.id)){ // NOPMD by wim on 9/20/10 1:39 PM
        return true;
      }else{
        return false;
      }
    } 
    if (properties.size() != e.properties.size()) {
      // logger.debug(" properties size " + properties.size() + " != " +
      // e.properties.size() );
      return false;
    }
    for (Property p : properties) {
      if (!p.equals(e.getProperty(p.getType()))) {
        // logger.debug("PROPERTIES");
        return false;
      }
    }
    if (this.references.size() == 0 && e.references.size() == 0) { // NOPMD JDK 1.5 by wim on 9/20/10 1:43 PM
      return true;
    }
    if (this.references.size() != e.references.size()) {
      // logger.debug(" refs size " + references.size() + " != " +
      // e.references.size() );
      return false;
    }
    /*
     * Iterator<ElementRefs> i1 = this.references.iterator();
     * Iterator<ElementRefs> i2 = e.references.iterator(); while (i1.hasNext())
     * { ElementRefs r1 = i1.next(); ElementRefs r2 = i2.next(); if
     * (!r1.equals(r2)) { return false; } }
     */
    return true;
  }

  /**
   * {@inheritDoc}
   */
  public int hashCode() {
    final int prime = 31; // NOPMD by wim on 9/20/10 1:39 PM
    int result = 1;
    for (Property p : properties) {
      result = prime * result + p.hashCode();
    }
    /*
     * watch out for recursion ! for (ElementRefs r : this.references) { result
     * = prime * result + r.hashCode(); }
     */
    return result;
  }

  public String toString() {
	final ToStringBuilder sb = new ToStringBuilder(this, // NOPMD by wim on 9/20/10 1:45 PM
        ToStringStyle.SHORT_PREFIX_STYLE);
    sb.append("id", this.id);
    sb.append("type", this.elementType.getName());
    sb.append("properties [");
    for (Property p : properties) {
      sb.append(p.toString());
    }
    sb.append("]");
    return sb.toString();
  }

  public void indent(final int ind) {
    System.out.print(ind + ">"); // NOPMD by wim on 9/20/10 1:44 PM
    for (int i = 0; i < ind; i++) {
      System.out.print("..."); // NOPMD by wim on 9/20/10 1:44 PM
    }
  }
  public void resetPrinting(){
    if( !printing ){
      return;
    }
    printing=false;
    for (ElementRefs r : this.references) {
      r.resetPrinting();
    }
  }
  public void print(final int ind, final String comment) {
    if( ind==0){
      resetPrinting();
    }
    indent(ind);
    System.out.println("BEGIN: =========> " + comment + " <========="); // NOPMD by wim on 9/20/10 1:44 PM
    indent(ind);
    final ToStringBuilder sb = new ToStringBuilder(this, // NOPMD by wim on 9/20/10 1:45 PM
        ToStringStyle.SHORT_PREFIX_STYLE);
    sb.append("id", this.id);
    sb.append("serialVersionUID", serialVersionUID);
    sb.append("version", version);
    sb.append("type", this.elementType.toString());
    System.out.println(sb.toString()); // NOPMD by wim on 9/20/10 1:44 PM
    if (!printing) {
      printing = true;
      for (Modification h : this.modifications) {
        h.print(ind + 1);
      }

      for (Property p : this.properties) {
        p.print(ind + 1);
      }

      for (ElementRefs r : this.references) {
        r.print(ind + 1);
      }
    }
    indent(ind);
    System.out.println("END: =========> " + comment + " <========="); // NOPMD by wim on 9/20/10 1:44 PM
  }
  
  /** assign element e to this element
   * 
   * @param from
   * @param knownElements 
   * @param elementDao
   * @throws CoreException
   */
  public void assign(Element from, List<Element> knownElements, ElementDao elementDao) throws CoreException {     
    logger.debug("begin of Element::assign()");

    assert (elementType != null && from.elementType != null);
    assert (elementType.equals(from.elementType));

    // assign properties
    for (Property propFrom : from.properties) {
      final Property propTo = getProperty(propFrom.getType()); 
      if (propTo != null) { 
    	  
        if (!propTo.equals(propFrom)) { 
        	logger.debug("assign modified Property {}", propFrom.getName());
        }
          propTo.assign(propFrom);
      }
      else {
        if (propFrom.isValid()) {
          propFrom.initRevision(getNewRevision());
          propFrom.setElement(this);
          logger.debug("add new Property {}", propFrom.getName());
          this.properties.add(propFrom);
        }
      }
    }

    final List<ElementRefs> fromRefsList = new ArrayList<ElementRefs>(from.references);
          
    // assign refs
    for (ElementRefs fromRef : fromRefsList) {
      ElementRefs toRefs = getElementRefs(fromRef.getRefName());
      if (toRefs == null) {
      	logger.debug("ElementRefs '{}' not available. Add new one", fromRef.getRefName());
        toRefs = new ElementRefs(this, fromRef);
      }
    }
    for (ElementRefs fromRef : fromRefsList) {
      ElementRefs toRefs = getElementRefs(fromRef.getRefName());
    	toRefs.assign(fromRef, knownElements, elementDao);
    }
    // search for refs missing in from 
    final Iterator<ElementRefs> ri = this.references.iterator(); // NOPMD by wim on 9/20/10 1:47 PM
    while( ri.hasNext()  ){
    	final ElementRefs r=ri.next(); // NOPMD by wim on 9/20/10 1:47 PM
      if( from.getElementRefs(r.getRefName()) == null ){ 
        r.assign(null, knownElements, elementDao);
      }
    }
  }
  //save myself
  public Element save(final ElementDao elementDao) throws CoreException {
    return elementDao.save(this);
  }

  // eliminates all invalid Properties and Elementrefs
  public boolean crunch() throws CoreException {
      logger.debug("begin of Element::crunch() {}", getTypeName());
    if (!crunched) {
      crunched = true;
      // Properties
      final Iterator<Property> pi = this.properties.iterator(); // NOPMD by wim on 9/20/10 1:49 PM
      while (pi.hasNext()) {
    	final Property p = pi.next(); // NOPMD by wim on 9/20/10 1:47 PM
        if (p.isntValid()) {
	    logger.debug("remove invalid property {}", p.getType().getName());
          pi.remove();
        }
      }
      // Element-References
      final Iterator<ElementRefs> ri = this.references.iterator(); // NOPMD by wim on 9/20/10 1:49 PM
      while (ri.hasNext()) {
    	final ElementRefs refs = ri.next();
        if (refs.crunch()) {
          ri.remove();
        }
      }
      logger.debug("end of crunch()");
    }
    return this.properties.isEmpty() && this.references.isEmpty();
  }

  public void delete() throws CoreException {
    if (isTransient()) {
      throw new CoreException("it is not allowed to delete a new Element");
    }
    final Modification last_mod = getLastModification();
    if (last_mod.getId() == null) {
      throw new CoreException(
          "it is not allowed to delete an already updated Element");
    }
    last_mod.setNextRevision(last_mod.getRevision() + 1);
  }

  // for debugging
  public boolean checkParent() {
    boolean wrong = false;
    for (Property p : this.properties) {
      if (p.checkParent(this)) {
	  logger.debug("Property {} with wrong Parent", p.getName());
        wrong = true;
      }
    }
    for (ElementRefs r : this.references) {
      if (r.checkParent(this)) {
	  logger.debug("Elementrefs {} with wrong Parent", r.getRefName());
        wrong = true;
      }
    }
    return wrong;
  }

  /**
   * get name of element type
   * @return name of element type
   */
  public String getTypeName() {
    return elementType.getName();
  }

  /**
   * get version (used for optimistic locking
   * @return version
   */
  public Long getVersion() {
    return version;
  }

  /**
   * set version (used for optimistic locking)
   */
  public void setVersion(final Long version) {
    this.version = version;
  }

  /**
   * mark this element as deleted 
   * @throws CoreException
   */
  public void setStateDeleted( ) throws CoreException {
  	Modification m = getNewModification();
  	m.setNextRevision(m.getRevision());
  }
  
  void setLoaded() {
    if (!isLoaded) {
      this.isLoaded = true;
      for (ElementRefs r : references) {
        r.setLoaded();
        logger.debug("loaded {}", r.getRefName());
      }
    }
  }

  boolean isLoaded() {
    return isLoaded;
  }
} // End of Element
